/*
 * DDR "configuration" class
 * 
 * References:
 *  - http://www.codeproject.com/csharp/codecompilation.asp
 *  - http://msdn2.microsoft.com/en-US/library/ms404279.aspx
 */

using System;
using System.IO;
using System.Reflection;
using DDRHelper.Services;
using System.Collections.Generic;
using System.Text;
using System.Windows.Forms;

namespace DDRHelper
{
    class DDRConfiguration
    {
        private List<DDRService> ddrservices;
        


        public DDRConfiguration()
        {
            ddrservices = new List<DDRService>();
        }

 

        public void loadConfiguration(String FileName)
        {
            if (!File.Exists(FileName))
            {
                if (FileName != Program.DefaultConfigurationFile)
                {
                    MessageBox.Show("Can't open the configuration file '" + FileName + "' because it doesn't exist.\r\n\r\nThe old or a blank configuration will be used instead.", "DDRHelper", MessageBoxButtons.OK, MessageBoxIcon.Exclamation);
                }
                else
                {
                    MessageBox.Show("Welcome to DDRHelper!\r\n\r\nAs this is probably the first time you've run DDRHelper, no default configuration file exists.  A blank one will be used instead.\r\n\r\nRemember to save a default configuration file during this session so that this message does not appear again.", "DDRHelper", MessageBoxButtons.OK, MessageBoxIcon.Information);
                }
                return;
            }
            
            StreamReader sr;
            try
            {
                sr = new StreamReader(FileName);
            }
            catch (Exception ex)
            {
                MessageBox.Show("There was a problem trying to load the configuration file '" + FileName + "': " + ex.GetType().Name, "DDRHelper", MessageBoxButtons.OK, MessageBoxIcon.Exclamation);
                return;
            }

            if (sr.ReadLine() != "DDRHelper")
            {
                // bad parse?
                MessageBox.Show("Cannot open configuration file - invalid parse.  Expected 'DDRHelper'.");
                throw new Exception("Invalid file format!");
            }

            DDRService ServiceObject;
            List<DDRService> newsvs = new List<DDRService>();

            while (!sr.EndOfStream)
            {
                try
                {
                    ServiceObject = DDRService.LoadFromFile(sr);
                    newsvs.Add(ServiceObject);
                }
                catch (Exception ex)
                {
                    MessageBox.Show("There was a problem loading the configuration file.  This probably is because a script hasn't been installed.\r\n\r\nIf you have loaded a new version of a script, try installing the old version again, then delete any services that use it then save the configuration and shutdown DDRHelper.  After that, install the new version and try again.\r\n\r\nDebug information:\r\n" + ex.ToString());
                    if (ddrservices == null)
                    {
                        ddrservices = new List<DDRService>();
                    }
                    return;
                }
            }

            ddrservices = newsvs;
        }

        public void saveConfiguration(String FileName)
        {
            try
            {
                StreamWriter sw = new StreamWriter(FileName);
                this.saveConfiguration(sw);
                sw.Flush();
                sw.Close();
            }
            catch (Exception ex)
            {
                MessageBox.Show("There was a problem saving the configuration to '" + FileName + "': " + ex.ToString());
            }
            return;
        }

        public void saveConfiguration(StreamWriter sw) {
            this.saveConfiguration(sw, "127.0.0.1");
        }

        public void saveConfiguration(StreamWriter sw, String ip)
        {
            sw.WriteLine("DDRHelper");
            foreach (DDRService server in ddrservices)
            {
                sw.WriteLine(server.GetType().Name);
                server.WriteToFile(sw, ip);
            }
            sw.Flush();
        }

        /*
         * We need to work around how the DDRService subclasses write to Streams,
         * and put it into a String so that the TalkBackComponent.
         * 
         * Ironically TalkBackComponent writes it to a Stream, this wraps it all
         * up.
         */
        public string dumpConfiguration(String ip)
        {
            Stream s = new MemoryStream(40980);
            StreamWriter sw = new StreamWriter(s);
            this.saveConfiguration(sw, ip);
            sw.Flush();
            s.Flush();
            s.Seek(0, SeekOrigin.Begin);
            byte[] buf = new byte[40980];
            s.Read(buf, 0, buf.Length);
            return Encoding.UTF8.GetString(buf, 0, buf.Length).TrimEnd('\0');
        }

        public void addService(DDRService service)
        {
            ddrservices.Add(service);
        }

        public void setServiceEnabledState(int index, bool enabled) {
            if ((ddrservices.Count > index) && (index >= 0)) {
                ddrservices[index].Enabled = enabled;
            }
        }

        public List<DDRService> dumpServices()
        {
            return ddrservices;
        }

        public DDRService getService(int index)
        {
            if ((ddrservices.Count > index) && (index >= 0))
            {
                return ddrservices[index];
            }
            else
            {
                throw new IndexOutOfRangeException("The service you selected is out of range!");
            }
        }

        public void setService(int index, DDRService service)
        {
            if ((ddrservices.Count > index) && (index >= 0))
            {
                ddrservices.RemoveAt(index);
                ddrservices.Insert(index, service);
            }
            else
            {
                throw new IndexOutOfRangeException("The service you selected is out of range!");
            }
        }

        public void deleteService(int index)
        {
            if ((ddrservices.Count > index) && (index >= 0))
            {
                ddrservices.RemoveAt(index);
            }
            else
            {
                throw new IndexOutOfRangeException("The service you selected is out of range!");
            }
        }

        public int getServiceCount()
        {
            return ddrservices.Count;
        }
    }
}
