using System;
using System.IO;
using System.Collections.Generic;
using System.Text;
using System.Reflection;
using System.Windows.Forms;
using System.Drawing;
using DDRHelper.Properties;

namespace DDRHelper.Services
{
    /// <summary>
    /// This class defines a basic, abstract, generic service.  It needs to have some sections defined before it
    /// is entirely usable.  Please see existing services as an example for writing new ones.
    /// </summary>
    public abstract class DDRService
    {
        protected String ServiceName;
        protected int ServicePort;
        protected String ServiceDescription;
        protected String ServiceType;
        protected int DefaultPort = 0;
        protected List<ServiceProperty> Properties;

        /// <summary>
        /// Wether the service is enabled or not.
        /// </summary>
        public bool Enabled = true;

        // Here we define some default values.

        /// <summary>
        /// An icon for the service, or null if you wish to use a default.
        /// </summary>
        public static Icon SERVICE_PLUGIN_ICON = System.Drawing.SystemIcons.Application;

        /// <summary>
        /// The name of the service plugin.
        /// </summary>
        public static String SERVICE_PLUGIN_NAME = "Unnamed Service";

        /// <summary>
        /// The version of the service plugin.
        /// </summary>
        public static String SERVICE_PLUGIN_VERSION = "1.0";

        /// <summary>
        /// The name of the author of the service plugin.
        /// </summary>
        public static String SERVICE_PLUGIN_AUTHOR = "Nobody";

        /// <summary>
        /// A Uri link to the website relating to the service plugin.
        /// </summary>
        public static Uri SERVICE_PLUGIN_URI = new Uri("http://www.example.com/");

        /// <summary>
        /// The default port that the service runs on.
        /// </summary>
        public static int SERVICE_DEFAULT_PORT = 1;

        /// <summary>
        /// The category of the service.
        /// </summary>
        public static String SERVICE_CATEGORY = "";

        /// <summary>
        /// The subcategory of the service.
        /// </summary>
        public static String SERVICE_SUBCATEGORY = "";


        /// <summary>
        /// All services should define a no-args constructor.  This is used when adding blank,
        /// new services and defines a simple service with reasonable defaults.
        /// </summary>
        public DDRService()
        {
            throw new NotImplementedException("You need to define a noargs constructor for your service containing default values.");
        }    

        /// <summary>
        /// This is a fullargs constructor, and generally you don't need to override this in child classes
        /// entirely.  It sets all of the parameters for you - all you need to do is set the DefaultPort
        /// property.
        /// </summary>
        /// <param name="ServiceName">The services name, for example, "My Web Server".</param>
        /// <param name="ServicePort">The port that the service runs on, for example, "80".</param>
        /// <param name="ServiceDescription">A description of the service, for example,
        /// "This web server has a couple of Flash games and movies on it".</param>
        /// <param name="ServiceType">This is a short service type name, for example, "HTTP".  Generally this
        /// hard coded in the child classes.</param>
        public DDRService(String ServiceName, int ServicePort, String ServiceDescription, String ServiceType)
        {
            this.Properties = new List<ServiceProperty>();
            this.ServiceName = ServiceName;
            this.ServiceDescription = ServiceDescription;
            this.ServiceType = ServiceType;
            this.InitProperties();
        }

        /// <summary>
        /// This method is used to initialize all non-standard properties of a service.  You'll need to define
        /// an implementation for this method in child classes.
        /// </summary>
        protected abstract void InitProperties();

        /// <summary>
        /// This method sets the service's port.  Generally only TCPService and UDPService need to define this,
        /// it's used for port number validation.  If your service doesn't allow it's port to be changed, or has
        /// some other strict rules regarding port use, then you should override this method.
        /// </summary>
        /// <param name="ServicePort">The service's new port number.</param>
        public abstract void setServicePort(int ServicePort);

        /// <summary>
        /// This method should return a Uri object, containing a hyperlink that points to the service.  This will
        /// be included in web links to the service.   Child classes will need to provide an implementation.
        /// </summary>
        /// <param name="address">The IP address of this system, as seen by the requesting host.  It is used in
        /// the construction of the URL.</param>
        /// <returns>A hyperlink to the service.</returns>
        public abstract Uri getResourceUri(String address);

        /// <summary>
        /// Gets the property at the selected index.
        /// </summary>
        /// <param name="index">The index of the property to retrieve.</param>
        /// <returns>The ServiceProperty at the index selected..</returns>
        public ServiceProperty getProperty(int index)
        {
            if ((Properties.Count > index) && (index >= 0))
            {
                return Properties[index];
            }
            else
            {
                throw new IndexOutOfRangeException("That property is out of range.");
            }
        }

        /// <summary>
        /// This is an internal method used to append properties, without checking if a similarly named property
        /// exists.
        /// </summary>
        /// <param name="prop">The property to add.</param>
        /// <seealso cref="setProperty(ServiceProperty)"/>
        private void addProperty(ServiceProperty prop)
        {
            Properties.Add(prop);
        }

        /// <summary>
        /// Find the index of the property with a specific label.
        /// </summary>
        /// <param name="Label">The label to search for.</param>
        /// <returns>The index of the service with the specified label, or -1 if it was not found.</returns>
        /// <seealso cref="getPropertyLabel(String)"/>
        public int getPropertyLabelIndex(String Label)
        {
            for (int i = 0; i < Properties.Count; i++)
            {
                if (Properties[i].getLabel().Equals(Label))
                {
                    return i;
                }
            }
            return -1;
        }

        /// <summary>
        /// Retrieve a property with a specified label.
        /// </summary>
        /// <param name="Label">The label to search for.</param>
        /// <returns>A ServiceProperty if the label was found, or null if none was found.</returns>
        public ServiceProperty getPropertyLabel(String Label)
        {
            int i = this.getPropertyLabelIndex(Label);
            if (i >= 0)
            {
                return this.getProperty(i);
            }
            else
            {
                return null;
            }
        }

        /// <summary>
        /// Attempt to reset a property to it's last or default value.
        /// </summary>
        /// <param name="index">The index of the property to reset.</param>
        public void resetProperty(int index)
        {
            if ((index >= 0) && (Properties.Count > index))
            {
                Properties[index].resetValue();
            }
            else
            {
                throw new IndexOutOfRangeException();
            }
        }

        /// <summary>
        /// An internal method for setting properties at an arbitrary index.
        /// </summary>
        /// <param name="index">The index of the property to set.</param>
        /// <param name="Property">A ServiceProperty to insert at the position.</param>
        /// <seealso cref="setProperty(ServiceProperty)"/>
        private void setPropertyAtIndex(int index, ServiceProperty Property)
        {
            if ((index >= 0) && (Properties.Count > index))
            {
                Properties[index] = Property;
            } else {
                throw new IndexOutOfRangeException();
            }
        }

        /// <summary>
        /// Sets the Property specified.  If it does not exist it will be added.  If it does exist then it shall
        /// replace an existing value.
        /// </summary>
        /// <param name="Property">The ServiceProperty to set.</param>
        public void setProperty(ServiceProperty Property)
        {
            int location;
            location = this.getPropertyLabelIndex(Property.getLabel());

            if (location >= 0)
            {
                this.setPropertyAtIndex(location, Property);
            }
            else
            {
                this.addProperty(Property);
            }
        }

        /// <summary>
        /// Gets the default port for the service instance.
        /// </summary>
        /// <returns>The service's default port.</returns>
        public int getDefaultPort()
        {
            return DefaultPort;
        }

        /// <summary>
        /// Gets the service's name.
        /// </summary>
        /// <returns>The service's name.</returns>
        public String getServiceName()
        {
            return ServiceName;
        }

        /// <summary>
        /// Gets the service's port.
        /// </summary>
        /// <returns>The service's port number.</returns>
        public int getServicePort()
        {
            return ServicePort;
        }

        /// <summary>
        /// Gets the service's type label.
        /// </summary>
        /// <returns>The service's type label.</returns>
        public String getServiceType()
        {
            return ServiceType;
        }

        /// <summary>
        /// Gets the service's description.
        /// </summary>
        /// <returns>The service's description.</returns>
        public String getServiceDescription()
        {
            return ServiceDescription;
        }

        /// <summary>
        /// Sets the service's description.
        /// </summary>
        /// <param name="ServiceDescription">The new value for the service's description.</param>
        public void setServiceDescription(String ServiceDescription)
        {
            this.ServiceDescription = ServiceDescription;
        }

        /// <summary>
        /// Set's the service's name.
        /// </summary>
        /// <param name="ServiceName">The new value for the service's name.</param>
        public void setServiceName(String ServiceName)
        {
            this.ServiceName = ServiceName;
        }


        /// <summary>
        /// Get a List of all properties relating to the service.
        /// </summary>
        /// <returns>A List of all ServiceProperties relating to the service.</returns>
        public List<ServiceProperty> dumpProperties()
        {
            return Properties;
        }

        /// <summary>
        /// Write a list of all properties to a file.
        /// </summary>
        /// <param name="sw">A StreamWriter handle to where the properties should be written to.</param>
        private void dumpProperties(StreamWriter sw)
        {
            for (int x = 0; x < this.Properties.Count; x++)
            {
                sw.WriteLine(this.Properties[x].GetType().Name);
                this.Properties[x].WriteToFile(sw);
            }
            sw.WriteLine("EndService");
        }

        /// <summary>
        /// Load a list of properties from a file.
        /// </summary>
        /// <param name="sr">A StreamReader handle to where the properties should be loaded from.</param>
        private void loadProperties(StreamReader sr)
        {
            String buf;
            List<ServiceProperty> lsp = new List<ServiceProperty>();
            buf = sr.ReadLine();
            Type PropertyType;
            Object[] param;
            while (buf != "EndService")
            {
                PropertyType = Assembly.GetEntryAssembly().GetType("DDRHelper.Properties." + buf, true, true);
                param = new Object[] { (Object)sr };
                lsp.Add((ServiceProperty)PropertyType.InvokeMember("LoadFromFile", BindingFlags.Public | BindingFlags.Static | BindingFlags.InvokeMethod, null, null, param));
                buf = sr.ReadLine();
            }
            Properties = lsp;
        }

        /// <summary>
        /// Loads a Service from a file.
        /// </summary>
        /// <param name="sr">A StreamReader handle that points to an open file.</param>
        /// <returns>A new DDRService-like object from the file.</returns>
        public static DDRService LoadFromFile(StreamReader sr)
        {
            DDRService newService;
            Type ServiceType;

            // use reflection to find the type that we want to create
            ServiceType = Program.servicesAssembly.GetType("DDRHelper.Services." + sr.ReadLine(), true, true);

            String ServiceName = sr.ReadLine();
            bool Enabled = Boolean.Parse(sr.ReadLine());
            //ignore the URL cache.
            sr.ReadLine();
            String ServiceDescription = sr.ReadLine().Replace("<br>", "\r\n");
            int ServicePort = Int32.Parse(sr.ReadLine());

            // invoke the specific type constructor using reflection
            Object[] args = { ServiceName, ServiceDescription, ServicePort };
            newService = (DDRService)ServiceType.InvokeMember("", BindingFlags.CreateInstance, null, null, args);

            newService.Enabled = Enabled;
            newService.loadProperties(sr);
            return newService;
        }

        /// <summary>
        /// Write the contents of the service to a file, using a default IP address in the caching of service
        /// URLs.
        /// </summary>
        /// <param name="sw">A StreamWriter handle pointing to the file to write to.</param>
        public void WriteToFile(StreamWriter sw)
        {
            this.WriteToFile(sw, "127.0.0.1");
        }

        /// <summary>
        /// Write the contents of the service to a file, using a specified IP address in the caching of service
        /// URLs.
        /// </summary>
        /// <param name="sw">A StreamWriter handle pointing to the file to write to.</param>
        /// <param name="ip">The IP address to use in the caching of service URLs.</param>
        public void WriteToFile(StreamWriter sw, String ip)
        {
            sw.WriteLine(this.ServiceName);
            sw.WriteLine(this.Enabled);
            sw.WriteLine(this.getResourceUri(ip));
            sw.WriteLine(this.ServiceDescription.Replace("\r\n", "<br>"));
            sw.WriteLine(this.ServicePort);
            this.dumpProperties(sw);
        }

        /// <summary>
        /// Present a form to the user allowing them to edit a property in a specialised manner.
        /// </summary>
        /// <param name="index">The index of the property to edit.</param>
        public void EditPropertyForm(int index)
        {
            Properties[index].EditPropertyForm();
        }

        /// <summary>
        /// Show an "about box" for the service.
        /// </summary>
        public static void showAboutDialogue()
        {
            MessageBox.Show("No about box defined!");
        }
    }
}
